var PORT = 9876
var PORTAND41 = PORT + 41
var PORTAND42 = PORT + 42
var PORTAND43 = PORT + 43
var PORTAND44 = PORT + 44
var PORTAND45 = PORT + 45
var PORTAND46 = PORT + 46
var PORTAND47 = PORT + 47
var PORTAND48 = PORT + 48
var PORTAND49 = PORT + 49
var PORTAND50 = PORT + 50
var PORTAND72 = PORT + 72
var PORTAND114 = PORT + 114
var PORTAND115 = PORT + 115
var PORTAND116 = PORT + 116
var PORTAND117 = PORT + 117
var PORTAND118 = PORT + 118
var PORTAND119 = PORT + 119
var PORTAND316 = PORT + 316
var PORTAND326 = PORT + 326
var PORTAND327 = PORT + 327

module.exports = {
  PORT,
  PORTAND41,
  PORTAND42,
  PORTAND43,
  PORTAND44,
  PORTAND45,
  PORTAND46,
  PORTAND47,
  PORTAND48,
  PORTAND49,
  PORTAND50,
  PORTAND72,
  PORTAND114,
  PORTAND115,
  PORTAND116,
  PORTAND117,
  PORTAND118,
  PORTAND119,
  PORTAND316,
  PORTAND326,
  PORTAND327
}
